(* Semantic checking for the Sipl compiler *)

open Ast

module StringMap = Map.Make(String)

(* Semantic checking of a program. Returns void if successful,
   throws an exception if something is wrong.

   Check each global variable, then check each function *)

let check (globals, functions) =

  (* Raise an exception if the given list has a duplicate *)
  let report_duplicate exceptf list =
    let rec helper = function
	n1 :: n2 :: _ when n1 = n2 -> raise (Failure (exceptf n1))
      | _ :: t -> helper t
      | [] -> ()
    in helper (List.sort compare list)
  in

  (* Raise an exception if a given binding is to a void type *)
  let check_not_void exceptf = function
      (Void, n) -> raise (Failure (exceptf n))
    | _ -> ()
  in
  
  (* Raise an exception of the given rvalue type cannot be assigned to
     the given lvalue type *)
  let check_assign lvaluet rvaluet err =
     if lvaluet == rvaluet || lvaluet == Float && rvaluet == Int then lvaluet else raise err
  in
   
  (**** Checking Global Variables ****)

  List.iter (check_not_void (fun n -> "illegal void global " ^ n)) globals;
   
  report_duplicate (fun n -> "duplicate global " ^ n) (List.map snd globals);

  (**** Checking Functions ****)

  if List.mem "print" (List.map (fun fd -> fd.fname) functions)
  then raise (Failure ("function print may not be defined")) else ();

  report_duplicate (fun n -> "duplicate function " ^ n)
    (List.map (fun fd -> fd.fname) functions);

  (* Function declaration for a named function *)

  let built_in_func_list = [
  { typ = Image; fname = "imgread"; formals = [(String, "x")]; locals = []; body = []};
  { typ = Void; fname = "imgwrite"; formals = [(Image, "img"); (Int, "x"); (String, "filename")]; locals = []; body = []};
  { typ = Image; fname = "imgCopy"; formals = [(Image, "img")]; locals = []; body = []};
  { typ = Image; fname = "imgCreate"; formals = [(Int, "x"); (Int, "x")]; locals = []; body = []};
  { typ = Matrix; fname = "matCreate"; formals = [(Int, "x"); (Int, "x")]; locals = []; body = []};
  { typ = Matrix; fname = "kernel"; formals = [(String, "x")]; locals = []; body = []};
  { typ = Float; fname = "int2float"; formals = [(Int, "x")]; locals = []; body = []};
  { typ = Float; fname = "power"; formals = [(Float, "x"); (Float, "y")]; locals = []; body = []};
  { typ = Int; fname = "float2int"; formals = [(Float, "x")]; locals = []; body = []};
  { typ = Image; fname = "convImg"; formals = [ (Image, "img"); (Matrix, "mat")]; locals = []; body = []};
  { typ = Void; fname = "changeGrey"; formals = [ (Image, "img") ]; locals = []; body = []};
  { typ = Void; fname = "rotateImage"; formals = [ (Image, "img") ]; locals = []; body = []};
  { typ = Void; fname = "flipImage"; formals = [ (Image, "img"); (Int, "x") ]; locals = []; body = []};
  { typ = Void; fname = "printMatrix"; formals = [(Matrix, "mat")]; locals = []; body = []};
  { typ = Void; fname = "printPixel"; formals = [(Pixel, "mat")]; locals = []; body = []};
  { typ = Void; fname = "prints"; formals = [(String, "x")]; locals = []; body = []};
  { typ = Void; fname = "printfloat"; formals = [(Float, "x")]; locals = []; body = []};
  { typ = Void; fname = "printImage"; formals = [(Image, "x")]; locals = []; body = []};
  { typ = Void; fname = "print"; formals = [(Int, "x")]; locals = []; body = []};
  { typ = Void; fname = "printb"; formals = [(Bool, "x")]; locals = []; body = []};
  { typ = Void; fname = "printbig"; formals = [(Int, "x")]; locals = []; body = []}
  ] in

  let built_in_decls =  List.fold_left (fun m fd -> StringMap.add fd.fname fd m)
     StringMap.empty built_in_func_list

   in
     
  let function_decls = List.fold_left (fun m fd -> StringMap.add fd.fname fd m)
                         built_in_decls functions
  in

  let function_decl s = try StringMap.find s function_decls
       with Not_found -> raise (Failure ("unrecognized function " ^ s))
  in

  let _ = function_decl "main" in (* Ensure "main" is defined *)

  let check_function func =

    List.iter (check_not_void (fun n -> "illegal void formal " ^ n ^
      " in " ^ func.fname)) func.formals;

    report_duplicate (fun n -> "duplicate formal " ^ n ^ " in " ^ func.fname)
      (List.map snd func.formals);

    List.iter (check_not_void (fun n -> "illegal void local " ^ n ^
      " in " ^ func.fname)) func.locals;

    report_duplicate (fun n -> "duplicate local " ^ n ^ " in " ^ func.fname)
      (List.map snd func.locals);

    (* Type of each variable (global, formal, or local *)
    let symbols = List.fold_left (fun m (t, n) -> StringMap.add n t m)
	StringMap.empty (globals @ func.formals @ func.locals )
    in

    let type_of_identifier s =
      try StringMap.find s symbols
      with Not_found -> raise (Failure ("undeclared identifier " ^ s))
    in

    (* Return the type of an expression or throw an exception *)
    let rec expr = function
	Literal _ -> Int
      | BoolLit _ -> Bool
      | StringLit _ -> String
      | FloatLit _ -> Float
      | MatrixLit l -> 
        let col = List.length (List.hd l) in
        let matched = List.for_all (fun li -> (List.length li) == col) l in 
      (match matched with
      | true -> Matrix
      | false -> raise (Failure ("Matrix Format Not Match"))
      )
      | MatAccess(s, r, c) -> let t0 = type_of_identifier s and t1 = expr r and t2 = expr c in
          (match t0 with
            Matrix when t1 = Int && t2 = Int -> Float
            | Matrix -> raise (Failure ("Matrix index should be Integer"))
            | _ -> raise (Failure ("identifier is not matrix"))
          )
      | MatAssign(s, r, c, e) -> 
              let t0 = type_of_identifier s and t1 = expr r 
                  and t2 = expr c and t3 = expr e in 
          (match t0 with
            Matrix when t1 = Int && t2 = Int && (t3 = Float || t3 = Int) -> Int
            | Matrix when t1 = Int && t2 = Int -> raise (Failure ("Matrix element value should be Float"))
            | Matrix -> raise (Failure ("Matrix index should be Integer"))
            | _ -> raise (Failure ("identifier is not matrix"))
          )
      | ImgAssign(s, r, c, ch, e) -> 
              let t0 = type_of_identifier s and t1 = expr r 
                  and t2 = expr c and t3 = expr e in 
          (match t0 with
            Image when t1 = Int && t2 = Int && t3 = Int -> 
              (match ch with
                | Id(_) -> Int
                | StringLit l when l = "r" || l = "g" || l = "b" -> Int
                | _ -> raise (Failure ("specified channel is not valid"))
              )
            | Image when t1 = Int && t2 = Int && t3 = Pixel -> 
            (match ch with
                | Id(_) -> Pixel
                | StringLit l when l = "a" -> Pixel
                | _ -> raise (Failure ("specified channel is not valid"))
              )
            | Image when t1 = Int && t2 = Int -> raise (Failure ("Image element value should be Integer"))
            | Image -> raise (Failure ("Image index should be Integer"))
            | _ -> raise (Failure ("identifier is not Image"))
          )
      | ImgAccess(s, r, c, ch) -> 
              let t0 = type_of_identifier s and t1 = expr r 
                  and t2 = expr c in 
          (match t0 with
            Image when t1 = Int && t2 = Int -> 
            (match ch with
                | Id(_) -> Int
                | StringLit l when l = "r" || l = "g" || l = "b" -> Int
                | StringLit l when l = "a" -> Pixel
                | _ -> raise (Failure ("specified channel is not valid"))
            )
            | Image -> raise (Failure ("Image index should be Integer"))
            | _ -> raise (Failure ("identifier is not Image"))
          )
      | GetWidth(id) -> let typ = type_of_identifier id in 
            (match typ with 
            Matrix -> Int
            | Image -> Int
            | _ -> raise(Failure("identifier is not matrix nor image"))
            ) 
      | GetHeight(id) -> let typ = type_of_identifier id in 
            (match typ with 
            Matrix -> Int
            | Image -> Int
            | _ -> raise(Failure("identifier is not matrix nor image"))
            ) 

      | Id s -> type_of_identifier s
      | Binop(e1, op, e2) as e -> let t1 = expr e1 and t2 = expr e2 in
	(match op with
          Add | Sub | Mult | Div when t1 = Int && t2 = Int -> Int
        | Add | Sub | Mult | Div when t1 = Float && (t2 = Int || t2 = Float) -> Float
        | Add | Sub | Mult | Div when t1 = Int && t2 = Float -> Float
        | Add | Div when t1 = Image && t2 = Int || t1 = Int && t2 = Image || t1 = Image && t2 = Image -> Image 
        | Add when t1 = Pixel && t2 = Int || t1 = Int && t2 = Pixel || t1 = Pixel && t2 = Pixel -> Pixel
        | Sub when t1 = Image && t2 = Int -> Image
        | Sub when t1 = Pixel && t2 = Int -> Pixel
        | Conv when t1 = Image && t2 = Matrix || t1 = Matrix && t2 = Image -> Image
        | Mult when t1 = Image && t2 = Float || t1 = Float && t2 = Image -> Image
        | Mult when t1 = Pixel && t2 = Float || t1 = Float && t2 = Pixel -> Pixel
	| Equal | Neq when t1 = t2 -> Bool
	| Less | Leq | Greater | Geq when (t1 = Int || t1 = Float) && (t2 = Float || t2 = Int) -> Bool
	| And | Or when t1 = Bool && t2 = Bool -> Bool
        | _ -> raise (Failure ("illegal binary operator " ^
              string_of_typ t1 ^ " " ^ string_of_op op ^ " " ^
              string_of_typ t2 ^ " in " ^ string_of_expr e))
        )
      | Unop(op, e) as ex -> let t = expr e in
	 (match op with
	   Neg when t = Int -> Int
   | Neg when t = Float -> Float
	 | Not when t = Bool -> Bool
         | _ -> raise (Failure ("illegal unary operator " ^ string_of_uop op ^
	  		   string_of_typ t ^ " in " ^ string_of_expr ex)))
      | Noexpr -> Void
      | Assign(var, e) as ex -> let lt = type_of_identifier var
                                and rt = expr e in
        check_assign lt rt (Failure ("illegal assignment " ^ string_of_typ lt ^
				     " = " ^ string_of_typ rt ^ " in " ^ 
				     string_of_expr ex))
      | Call(fname, actuals) as call -> let fd = function_decl fname in
         if List.length actuals != List.length fd.formals then
           raise (Failure ("expecting " ^ string_of_int
             (List.length fd.formals) ^ " arguments in " ^ string_of_expr call))
         else
           List.iter2 (fun (ft, _) e -> let et = expr e in
              ignore (check_assign ft et
                (Failure ("illegal actual argument found " ^ string_of_typ et ^
                " expected " ^ string_of_typ ft ^ " in " ^ string_of_expr e))))
             fd.formals actuals;
           fd.typ
    in

    let check_bool_expr e = if expr e != Bool
     then raise (Failure ("expected Boolean expression in " ^ string_of_expr e))
     else () in

    (* Verify a statement or throw an exception *)
    let rec stmt = function
	Block sl -> let rec check_block = function
           [Return _ as s] -> stmt s
         | Return _ :: _ -> raise (Failure "nothing may follow a return")
         | Block sl :: ss -> check_block (sl @ ss)
         | s :: ss -> stmt s ; check_block ss
         | [] -> ()
        in check_block sl
      | Expr e -> ignore (expr e)
      | Return e -> let t = expr e in if t = func.typ then () else
         raise (Failure ("return gives " ^ string_of_typ t ^ " expected " ^
                         string_of_typ func.typ ^ " in " ^ string_of_expr e))
           
      | If(p, b1, b2) -> check_bool_expr p; stmt b1; stmt b2
      | For(e1, e2, e3, st) -> ignore (expr e1); check_bool_expr e2;
                               ignore (expr e3); stmt st
      | While(p, s) -> check_bool_expr p; stmt s
    in

    stmt (Block func.body)
   
  in
  List.iter check_function functions
